class
	CODE_TRANSFORMER

inherit
	BINDINGS_STATUS

create
	make

feature {NONE}

	make (a_project_settings: PROJECT_SETTINGS)
		require
			a_project_settings /= Void
			a_project_settings.cil_file /= Void
			a_project_settings.output_dir /= Void
		do
			project_settings := a_project_settings
			steps := <<
					-- create tree:
					create {SOURCE_LOADER},

					-- basic code cleanup:
					create {TYPEDEF_REMOVER},
					create {ENUM_TYPE_TRANSFORMER},

					-- add bindings:
					create {FORWARD_DECLARATIONS_REMOVER},
					create {BINDING_ADDER},

					-- do transformations:
					create {INCREMENT_DECREMENT_TRANSFORMER},
					create {SIZEOF_TRANSFORMER},
					create {FLUSH_TRANSFORMER},
					create {CIL_CHAR_TRANSFORMER},
					create {CIL_POINTER_ACCESS_TRANSFORMER},
					create {ASSIGNMENT_TRANSFORMER},
					create {POINTER_ACCESS_INDEX_TRANSFORMER},
					create {ARITHMETIC_OPERATION_TRANSFORMER},
					create {CAST_REMOVER},
					create {POINTER_TRANSFORMER},
					create {CAST_TRANSFORMER},
					create {PRINTF_TRANSFORMER},
					create {STD_STREAM_TRANSFORMER},
					create {LITERAL_TRANSFORMER},
					create {BOOLEAN_TRANSFORMER},
					create {PARAMETER_ASSIGNMENT_TRANSFORMER},
					create {ASSEMBLER_TRANSFORMER},
					create {VARIABLE_ADDRESS_TRANSFORMER},
					create {ADDRESS_OPERATOR_TRANSFORMER},
					create {IDENTIFIER_TRANSFORMER},
					create {BREAK_CONTINUE_RETURN_TRANSFORMER},
					create {LONGJMP_TRANSFORMER},
					create {SWITCH_TRANSFORMER},
					create {JUMP_STATEMENT_TRANSFORMER},
					create {VARARG_TRANSFORMER},
					create {ALLOCA_TRANSFORMER},
					create {BUILTIN_TRANSFORMER},
					create {EXPRESSION_AS_INSTRUCTION_TRANSFORMER},
					create {EXPANDED_TRANSFORMER},
					create {STRUCT_TRANSFORMER},

					-- O-O reengineering
					create {CFG_CREATOR},
					create {OO_START},
					create {OO_INHERITANCE_TRANSFORMER},
					create {OO_FUNCTIONS_TRANSFORMER},
					create {OO_CONTRACT_ADDER},
					create {OO_FINISH},

					-- final cleanup:
					create {CIL_TEMP_VARIABLE_REMOVER},

					-- emit Eiffel source code:
					create {EIFFEL_FILE_WRITER}
				>>
			current_step_index := 0
			progress_listener := create {TWO_WAY_LIST [PROCEDURE [ANY, TUPLE [INTEGER]]]}.make
			create tree.make
		ensure
			project_settings = a_project_settings
			steps /= Void and current_step_index = 0
		end

feature {ANY}

	project_settings: PROJECT_SETTINGS

	steps: ARRAY [TRANSFORMATION_STEP]

	current_step_index: INTEGER

	progress_listener: LIST [PROCEDURE [ANY, TUPLE [INTEGER]]]

	go (target_step_index: INTEGER): BOOLEAN
		require
			target_step_index >= 1 and target_step_index <= steps.count
		local
			done, ok: BOOLEAN
		do
			ok := True
			from
				if target_step_index <= current_step_index then
					current_step_index := 1
				else -- target_step_index > current_step_index
					current_step_index := current_step_index + 1
				end
			until
				done
			loop
				if current_step_index = 1 then
					Bindings_active.set_item (False) -- deactivate binding checking
				end
				notify_progress_listener
				ok := steps.item (current_step_index).process (Current)
				if not ok then
					done := True
					current_step_index := 0
					notify_progress_listener
				elseif current_step_index = target_step_index then
					done := True
				else
					current_step_index := current_step_index + 1
				end
			end
			Result := ok
		end

	emit_tree (ev_tree: EV_TREE)
		require
			current_step_index > 0
		do
			steps.item (current_step_index).emit_tree (ev_tree)
		end

	emit_source (text: EV_TEXT)
		require
			current_step_index > 0
		do
			steps.item (current_step_index).emit_source (text)
		end

feature {ANY}

	tree: AST_TRANSLATION_UNIT

feature {NONE}

	notify_progress_listener
		do
			across progress_listener as c loop
				c.item.call ([current_step_index])
			end
		end

invariant
	tree /= Void

end
