class
	AST_VARIABLE_DECLARATION_OR_TYPE_HELPER

inherit
	ANTLR_TRANSFORMATION_UTILITIES

create
	make_from_antlr

feature {NONE}

	make_from_antlr (node: ANTLR_AST_NODE)
		require
			node /= Void and then node.type = Ast_variable_declaration_or_type
		local
			cur_head: ANTLR_AST_NODE
			cur_parameter_list: AST_PARAMETER_LIST
		do
			-- storage_class_specifier*
			node.children.start
			from
				storage_class_specifiers := create {TWO_WAY_LIST [INTEGER]}.make
			until
				node.children.item.type /= Ast_storage_class_specifier
			loop
				storage_class_specifiers.extend (node.children.item.children.first.type)
				node.children.forth
			end

			-- type
			type := process_type (node.children.item, Void)
			node.children.forth

			-- attribute_specifier?
			if
				not node.children.after and then
				node.children.item.type = C_kw_attribute
			then
				update_gcc_attribute_specifier (node.children.item)
				node.children.forth
			end

			if not node.children.after then
				if node.children.item.type = Ast_pointer_type then
					-- pointer_type IDENTIFIER? array_type? array_type
					type := create {AST_POINTER_TYPE}.make_from_antlr (
							node.children.item,
							create {AST_ARRAY_TYPE}.make_from_antlr (node.children.last, type, Void),
							Void)
					node.children.forth
					if node.children.item.type /= Ast_array_type then
						identifier := node.children.item.token_text
						node.children.forth
					end
					if node.children.item /= node.children.last then
						type := create {AST_ARRAY_TYPE}.make_from_antlr (node.children.item, type, Void)
					end
				else
					-- variable_declaration_head
					from
						cur_head := node.children.item
					until
						cur_head = Void
					loop
						inspect
							cur_head.children.first.type
						when Ast_pointer_type, C_kw_attribute then
							-- do that first (before updating `type')
							if cur_head.children.last.type = Ast_parameter_list then
								cur_parameter_list := create {AST_PARAMETER_LIST}.make_from_antlr (cur_head.children.last, Void)
							else
								cur_parameter_list := create {AST_PARAMETER_LIST}.make
							end

							cur_head.children.start
							if cur_head.children.item.type = C_kw_attribute then
								update_gcc_attribute_specifier (cur_head.children.item)
								cur_head.children.forth
							end
							type := create {AST_POINTER_TYPE}.make_from_antlr (
									cur_head.children.item,
									create {AST_FUNCTION_TYPE}.make (type, cur_parameter_list),
									Void
								)
							cur_head.children.forth
							if
								not cur_head.children.after and then
								cur_head.children.item.type = Ast_variable_declaration_or_type_head
							then
								cur_head := cur_head.children.item
							else
								cur_head := Void
							end
						when Ast_array_type then
							type := create {AST_ARRAY_TYPE}.make_from_antlr (cur_head.children.first, type, Void)
							cur_head := Void
						else
							identifier := cur_head.children.first.token_text
							-- array_type?
							if cur_head.children.last.type = Ast_array_type then
								type := create {AST_ARRAY_TYPE}.make_from_antlr (cur_head.children.last, type, Void)
							end
							cur_head := Void
						end
					end
					node.children.forth

					-- attribute_specifier?
					if
						not node.children.after and then
						node.children.item.type = C_kw_attribute
					then
						update_gcc_attribute_specifier (node.children.item)
						node.children.forth
					end

					-- (assignment_expression | array_initializer)?
					if not node.children.after then
						initializer := process_expression (node.children.item, Void)
					end
				end
			end
		end

feature {ANY}

	is_type: BOOLEAN
		do
			Result := identifier = Void
		end

	is_variable_declaration: BOOLEAN
		do
			Result := identifier /= Void
		end

	create_type: AST_TYPE
		require
			is_type
		do
			Result := type
		ensure
			Result /= Void and then Result.parent = Void
		end

	create_variable_declaration: AST_VARIABLE_DECLARATION
		require
			is_variable_declaration
		do
			Result := create {AST_VARIABLE_DECLARATION}.make (type, identifier, initializer)
			across storage_class_specifiers as c loop
				Result.storage_class_specifiers.extend (c.item)
			end
		ensure
			Result /= Void and then Result.parent = Void
		end

feature {AST_VARIABLE_DECLARATION}

	identifier: STRING

	storage_class_specifiers: LIST [INTEGER]

	type: AST_TYPE

	gcc_attribute_specifier: AST_GCC_ATTRIBUTE_SPECIFIER

	initializer: AST_EXPRESSION

feature {NONE}

	update_gcc_attribute_specifier (a_node: ANTLR_AST_NODE)
		require
			a_node /= Void and then a_node.type = C_kw_attribute
		do
			if gcc_attribute_specifier = Void then
				create gcc_attribute_specifier.make_from_antlr (a_node, Void)
			else
				gcc_attribute_specifier.merge (create {AST_GCC_ATTRIBUTE_SPECIFIER}.make_from_antlr (a_node, Void))
			end
		end

invariant
	is_type /= is_variable_declaration
	is_type implies (identifier = Void and storage_class_specifiers.is_empty and initializer = Void)
	is_variable_declaration implies (identifier /= Void)

	identifier /= Void implies not identifier.is_empty
	storage_class_specifiers /= Void
	type /= Void

end
