class
	AST_PARAMETER_LIST

inherit
	AST_NODE

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make
		do
			parameters := create {TWO_WAY_LIST [AST_NODE]}.make
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_parameter_list
		local
			l_h: AST_VARIABLE_DECLARATION_OR_TYPE_HELPER
			l_type: AST_TYPE
		do
			parent := init_parent

			parameters := create {TWO_WAY_LIST [AST_NODE]}.make
			across node.children as c loop
				inspect
					c.item.type
				when Ast_variable_declaration_or_type then
					l_h := create {AST_VARIABLE_DECLARATION_OR_TYPE_HELPER}.make_from_antlr (c.item)
					if l_h.is_type then
						l_type := l_h.create_type
						if attached {AST_BASIC_TYPE} l_type as bt and then bt.type = Ast_bt_void then
							-- ignore it; foo(void)  ->  foo()
						else
							parameters.extend (l_type)
							parameters.last.parent := Current
						end
					else
						parameters.extend (l_h.create_variable_declaration)
						parameters.last.parent := Current
					end
				when C_op_three_points then
					is_vararg := True
				end
			end
		end

feature {ANY}

	parameters: LIST [AST_NODE]

	is_vararg: BOOLEAN

	insert (a_new_parameter: AST_NODE; a_index: INTEGER_32)
			-- insert `a_new_parameter' at index `a_index' shifting all parameters starting at that index by one
		require
			attached {AST_VARIABLE_DECLARATION} a_new_parameter or attached {AST_TYPE} a_new_parameter
			a_new_parameter.parent = Void
			a_index >= 1 and a_index <= parameters.count + 1
		do
			if a_index > parameters.count then
				parameters.extend (a_new_parameter)
			else
				parameters.go_i_th (a_index)
				({TWO_WAY_LIST [AST_NODE]} #? parameters).put_left (a_new_parameter)
			end
			a_new_parameter.parent := Current
		ensure
			parameters.i_th (a_index) = a_new_parameter
			parameters.count = old parameters.count + 1
			a_new_parameter.parent = Current
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_parameter_list (Current)
			if visit_children then
				visit_list (parameters, v)
			end
			v.end_visit_parameter_list (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			parameters := create {TWO_WAY_LIST [AST_NODE]}.make
			across other.parameters as c loop
				parameters.extend (c.item.ast_twin)
				parameters.last.parent := Current
			end
			is_vararg := other.is_vararg
		end

invariant
	parameters /= Void and then across parameters as c all attached {AST_VARIABLE_DECLARATION} c.item or attached {AST_TYPE} c.item end

end
