class
	STRUCT_WRITER_OO

inherit
	NAME_UTILITIES

create
	make

feature {NONE}

	make (a_node: AST_STRUCT_DECLARATION)
		require
			a_node /= Void
		do
			node := a_node
		end

feature {NONE}

	node: AST_STRUCT_DECLARATION

feature {NONE} -- convenience data

	fields: LIST [AST_STRUCT_FIELD]
		-- the fields defined in the class (without the inherited ones)

	static_class_needs_asm, regular_class_needs_asm: BOOLEAN

	is_union, has_bitfields, needs_initializers: BOOLEAN
		-- we need "make" if any of those 3 is True

feature {ANY}

	buffer: STRING_8

	static_buffer: STRING_8

	write
		local
			l_data: CELL [BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			-- set `static_class_needs_asm' and `regular_class_needs_asm'
			create l_data.put (False)
			create l_visitor
			l_visitor.agents.end_visit_eiffel_expression := agent (ag_node: AST_EIFFEL_EXPRESSION; ag_data: CELL [BOOLEAN])
				do
					if
						attached {STRING_8} ag_node.mixed.first as s and then
						s.count >= 4 and then s.starts_with ("asm") and then
						s[4] >= '0' and then s[4] <= '9'
					then
						ag_data.put (True)
					end
				end (?, l_data)
			across
				node.oo_static_features as c
			until
				l_data.item
			loop
				c.item.accept (l_visitor)
			end
			static_class_needs_asm := l_data.item
			if not static_class_needs_asm then
				across
					node.oo_features as c
				until
					l_data.item
				loop
					c.item.accept (l_visitor)
				end
				regular_class_needs_asm := l_data.item
			end

			-- write the classes
			write_class
			write_static_class
		end

feature {NONE}

	write_class
		do
			if
				not node.source_info.is_pure_oo or
				not node.oo_features.is_empty
			then
				create buffer.make_empty

				if not node.source_info.is_pure_oo then
					fields := node.oo_defined_fields

					is_union := node.is_union
					has_bitfields := across fields as c some
							attached {AST_BITFIELD} c.item
						end
					needs_initializers := across fields as c some
							attached {AST_POINTER_TYPE} c.item.type or
							attached {AST_STRUCT_TYPE} c.item.type or
							attached {AST_ARRAY_TYPE} c.item.type
						end
				end

				buffer.append ("class%N")
				buffer.append ("	" + eiffel_class_name (node.identifier) + "%N%N")

				write_inherit_clause
				write_create_clause

				write_make
				write_struct_fields
				write_struct_field_setters

				write_features (buffer, node.oo_features)

				write_copy_is_equal
				write_structure_info
				write_bitfield_helper

				buffer.append ("end%N")
			end
		end

	write_static_class
		local
			l_first: BOOLEAN
		do
			if
				not node.oo_static_features.is_empty or
				across node.oo_features as c some attached {AST_FUNCTION_DECLARATION} c.item as fd and then fd.needs_address end
			then
				create static_buffer.make_empty

				static_buffer.append ("class%N")
				static_buffer.append ("	" + eiffel_class_name (node.identifier) + "_STATIC%N%N")

				static_buffer.append ("inherit%N")
				static_buffer.append ("	CE_UTILITIES%N")
				across node.oo_static_superclasses as c loop
					static_buffer.append ("	" + eiffel_class_name (c.item.identifier) + "_STATIC%N")
				end
				if static_class_needs_asm then
					static_buffer.append ("	" + eiffel_class_name (({AST_TRANSLATION_UNIT} #? node.parent).source_file_name) + "_ASSEMBLER%N")
				end
				static_buffer.append ("%N")

				write_features (static_buffer, node.oo_static_features)

				-- print required routine addresses
				l_first := True
				across node.oo_static_features as c loop
					if attached {AST_FUNCTION_DECLARATION} c.item as fd and then fd.needs_address then
						if l_first then
							static_buffer.append ("feature {ANY} -- routine addresses%N%N")
							l_first := False
						end
						write_routine_address (fd)
					end
				end
				across node.oo_features as c loop
					if attached {AST_FUNCTION_DECLARATION} c.item as fd and then fd.needs_address then
						if l_first then
							static_buffer.append ("feature {ANY} -- routine addresses%N%N")
							l_first := False
						end
						write_routine_address (fd)
					end
				end

				static_buffer.append ("end%N")
			end
		end

	write_routine_address (a_node: AST_FUNCTION_DECLARATION)
		local
			l_orig_params: AST_PARAMETER_LIST
			l_has_result: BOOLEAN
		do
			static_buffer.append ("	" + a_node.identifier + "_address: CE_ROUTINE%N")
			static_buffer.append ("		once (%"PROCESS%")%N")

			if a_node.oo_class.oo_static_features.has (a_node) then
				--> address of static feature
				if a_node.is_extern and a_node.parameter_list.is_vararg then
					-- Use the real C function pointer for external vararg functions. They don't
					-- work with a bridge like: extern_c() -> eif_wrapper(...) -> extern_c(...)
					static_buffer.append ("			create Result.make_shared (c_" + a_node.identifier + ")%N")
				else
					static_buffer.append ("			create Result.make_agent (agent " + a_node.identifier + ")%N")
				end
			elseif a_node.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_struct_pointer_arg then
				--> address of member feature with extracted argument
				l_orig_params := a_node.parameter_list.ast_twin
				l_orig_params.insert (
						a_node.oo_class_assignment_data.removed_parameter.ast_twin,
						a_node.oo_class_assignment_data.removed_parameter_index
					)

				if a_node.is_extern and a_node.parameter_list.is_vararg then
					static_buffer.append ("TODO?!") -- forgot what's that about (see above)
				else
					l_has_result := not (attached {AST_BASIC_TYPE} a_node.return_type as bt and then bt.type = Ast_bt_void)
					static_buffer.append ("			create Result.make_agent (agent (" + l_orig_params.eiffel_code + ")")
					if l_has_result then
						static_buffer.append (": " + a_node.return_type.eiffel_code)
					end
					static_buffer.append ("%N")
					static_buffer.append ("				do%N")
					static_buffer.append ("					")
					if l_has_result then
						static_buffer.append ("Result := ")
					end
					static_buffer.append (({AST_VARIABLE_DECLARATION} #? a_node.oo_class_assignment_data.removed_parameter).identifier)
					if a_node.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_struct_pointer_arg then
						static_buffer.append (".item")
					end
					static_buffer.append ("." + a_node.identifier)
					if not a_node.parameter_list.parameters.is_empty then
						static_buffer.append (" (")
						across a_node.parameter_list.parameters as pc loop
							if pc.target_index > 1 then
								static_buffer.append (", ")
							end
							static_buffer.append (({AST_VARIABLE_DECLARATION} #? pc.item).identifier)
						end
						static_buffer.append (")")
					end
					static_buffer.append ("%N")
					static_buffer.append ("				end)%N")
				end
			else
				--> address of member feature without extracted argument; it's kinda static -> create a dummy object
				if a_node.is_extern and a_node.parameter_list.is_vararg then
					static_buffer.append ("TODO?!") -- forgot what's that about (see above)
				else
					static_buffer.append ("			create Result.make_agent (agent (create {" + eiffel_class_name (node.identifier) +  "}.make)." + a_node.identifier + ")%N")
				end
			end

			static_buffer.append ("		end%N%N")
		end

feature {NONE}

	write_inherit_clause
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("inherit%N")
				if
					not node.oo_static_features.is_empty or
					across node.oo_features as c some attached {AST_FUNCTION_DECLARATION} c.item as fd and then fd.needs_address end
				then
					buffer.append ("	" + node.identifier + "_STATIC%N")
				else
					buffer.append ("	CE_UTILITIES%N")
				end
				buffer.append ("		redefine%N")
				buffer.append ("			copy,%N")
				buffer.append ("			is_equal%N")
				buffer.append ("		end%N")
				if
					node.oo_superclasses.is_empty or
					across node.oo_superclasses as c all c.item.source_info.is_pure_oo end
				then
					buffer.append ("	CE_CLASS%N")
					buffer.append ("		redefine%N")
					if is_union or has_bitfields or needs_initializers then
						buffer.append ("			make,%N")
					end
					buffer.append ("			copy,%N")
					buffer.append ("			is_equal")
					if has_bitfields then
						buffer.append (",%N")
						buffer.append ("			bitfield_helper%N")
					else
						buffer.append ("%N")
					end
					buffer.append ("		end%N")
					across node.oo_superclasses as c loop
						buffer.append ("	" + c.item.identifier + "_STATIC%N")
						buffer.append ("		redefine%N")
						buffer.append ("			copy,%N")
						buffer.append ("			is_equal%N")
						buffer.append ("		end%N")
					end
				else
					across node.oo_superclasses as c loop
						buffer.append ("	" + c.item.identifier)
						if c.item.source_info.is_pure_oo then
							buffer.append ("_STATIC%N")
							buffer.append ("		redefine%N")
							buffer.append ("			copy,%N")
							buffer.append ("			is_equal%N")
							buffer.append ("		end%N")
						else
							buffer.append ("%N")

							-- rename
							l_first := True
							node.fields.start
							across c.item.fields as c2 loop
								if
									attached c2.item.identifier as super_identifier and then
									attached node.fields.item.identifier as sub_identifier and then
									not super_identifier.is_equal (sub_identifier)
								then
									if l_first then
										buffer.append ("		rename%N")
										l_first := False
									else
										buffer.append (",%N")
									end
									buffer.append ("			" + super_identifier + " as " + sub_identifier)
								end
								node.fields.forth
							end
							if not l_first then
								buffer.append ("%N")
							end

							-- redefine
							buffer.append ("		redefine%N")
							if is_union or has_bitfields or needs_initializers then
								buffer.append ("			make,%N")
							end
							buffer.append ("			copy,%N")
							buffer.append ("			is_equal,%N")
							buffer.append ("			structure_info")
							if has_bitfields then
								buffer.append (",%N")
								buffer.append ("			bitfield_helper%N")
							else
								buffer.append ("%N")
							end

							buffer.append ("		end%N")
						end
					end
				end
				if regular_class_needs_asm then
					buffer.append ("	" + eiffel_class_name (({AST_TRANSLATION_UNIT} #? node.parent).source_file_name) + "_ASSEMBLER%N")
					buffer.append ("		redefine%N")
					buffer.append ("			copy,%N")
					buffer.append ("			is_equal%N")
					buffer.append ("		end%N")
				end
				buffer.append ("%N")
			end
		end

	write_create_clause
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("create%N")
				buffer.append ("	make,%N")
				buffer.append ("	make_shared,%N")
				buffer.append ("	make_initialized%N%N")
			end
		end

	write_make
		local
			l_tmp_array: AST_ARRAY_TYPE
		do
			if
				not node.source_info.is_pure_oo and then
				(is_union or has_bitfields or needs_initializers)
			then
				buffer.append ("feature {NONE}%N%N")
				buffer.append ("	make%N")
				buffer.append ("		do%N")
				if is_union or has_bitfields then
					buffer.append ("			make_memory -- ")
					if is_union and has_bitfields then
						buffer.append ("union and has bitfields")
					elseif is_union then
						buffer.append ("union")
					else
						buffer.append ("has bitfields")
					end
					buffer.append ("; back with memory for correct semantics%N")
				else
					across fields as c loop
						if
							attached {AST_POINTER_TYPE} c.item.type or
							attached {AST_ARRAY_TYPE} c.item.type or
							attached {AST_STRUCT_TYPE} c.item.type
						then
							buffer.append ("			create ")
							if
								(attached {AST_POINTER_TYPE} c.item.type as pt and then not attached {AST_FUNCTION_TYPE} pt.type) or
								attached {AST_ARRAY_TYPE} c.item.type
							then
								buffer.append ("{" + effective_type (c.item.type) + "} ")
							end
							if attached {AST_ARRAY_TYPE} c.item.type as a then
								buffer.append (c.item.identifier + ".make (<<")
								from
									l_tmp_array := a
								until
									l_tmp_array = Void
								loop
									if l_tmp_array /= a then
										buffer.append (", ")
									end
									if l_tmp_array.dimension /= Void then
										buffer.append (l_tmp_array.dimension.eiffel_code)
									else
										-- flexible array member at end of struct; no dimension needed
										buffer.append ("0")
									end
									l_tmp_array ?= l_tmp_array.type
								end
								buffer.append (">>)%N")
							else
								buffer.append (c.item.identifier + ".make%N")
							end
						end
					end
					if not node.oo_superclasses.is_empty then
						buffer.append ("			Precursor%N")
					end
				end
				buffer.append ("		end%N%N")
			end
		end

	write_struct_fields
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				l_first := True
				across fields as c loop
					if attached c.item.identifier as identifier then
						if l_first then
							buffer.append ("feature {ANY} -- fields%N%N")
							l_first := False
						end

						buffer.append ("	" + identifier + ": ")
						buffer.append (c.item.type.eiffel_code)
						if not attached {AST_ARRAY_TYPE} c.item.type then
							buffer.append (" assign " + create_assigner_name (node, c.item))
						end
						buffer.append ("%N%N")
					end
				end
			end
		end

	write_struct_field_setters
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				l_first := True
				across fields as c loop
					if
						attached c.item.identifier as identifier and
						not attached {AST_ARRAY_TYPE} c.item.type
					then
						if l_first then
							buffer.append ("feature {ANY} -- field setters%N%N")
							l_first := False
						end

						buffer.append ("	" + create_assigner_name (node, c.item) + " (a_" + identifier + ": ")
						buffer.append (c.item.type.eiffel_code)
						buffer.append (")%N")
						buffer.append ("		do%N")
						buffer.append ("			" + identifier + " := a_" + identifier + "%N")
						buffer.append ("			update_memory_field (%"" + identifier + "%")%N")
						buffer.append ("		end%N%N")
					end
				end
			end
		end

	write_copy_is_equal
		local
			l_first: BOOLEAN
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("feature {ANY}%N%N")

				buffer.append ("	copy (a_other: like Current)%N")
				buffer.append ("		do%N")
				across node.oo_superclasses as c loop
					if not c.item.source_info.is_pure_oo then
						buffer.append ("			Precursor {" + c.item.identifier + "} (a_other)%N")
					end
				end
				across fields as c loop
					if attached c.item.identifier as identifier then
						buffer.append ("			" + identifier + " := a_other." + identifier)
						if
							attached {AST_STRUCT_TYPE} c.item.type or
							attached {AST_ARRAY_TYPE} c.item.type
						then
							buffer.append (".twin")
						end
						buffer.append ("%N")
					end
				end
				if is_union or has_bitfields then
					buffer.append ("			create_memory_copy -- ")
					if is_union and has_bitfields then
						buffer.append ("union and has bitfields")
					elseif is_union then
						buffer.append ("union")
					else
						buffer.append ("has bitfields")
					end
					buffer.append ("; back with memory for correct semantics%N")
				end
				buffer.append ("		end%N%N")

				buffer.append ("	is_equal (a_other: like Current): BOOLEAN%N")
				buffer.append ("		do%N")
				buffer.append ("			Result :=%N")
				if
					node.oo_superclasses.is_empty and
					(
						fields.is_empty or else
						across fields as c all c.item.identifier = Void end
					)
				then
					buffer.append ("True%N")
				else
					l_first := True
					across node.oo_superclasses as c loop
						if not c.item.source_info.is_pure_oo then
							if l_first then
								l_first := False
							else
								buffer.append (" and%N")
							end
							buffer.append ("				Precursor {" + c.item.identifier + "} (a_other)")
						end
					end
					across fields as c loop
						if attached c.item.identifier as identifier then
							if l_first then
								l_first := False
							else
								buffer.append (" and%N")
							end
							buffer.append ("				")
							if
								attached {AST_POINTER_TYPE} c.item.type or
								attached {AST_STRUCT_TYPE} c.item.type or
								attached {AST_ARRAY_TYPE} c.item.type
							then
								buffer.append (identifier + ".is_equal (a_other." + identifier + ")")
							else
								buffer.append (identifier + " = a_other." + identifier)
							end
						end
					end
					buffer.append ("%N")
				end
				buffer.append ("		end%N%N")
			end
		end

	write_structure_info
		local
			l_first: BOOLEAN
			l_tmp_array: AST_ARRAY_TYPE
		do
			if not node.source_info.is_pure_oo then
				buffer.append ("feature {NONE}%N%N")
				buffer.append ("	structure_info: CE_STRUCTURE%N")
				buffer.append ("		once (%"PROCESS%")%N")

				if is_union then
					buffer.append ("			create Result.make_union (<<%N")
				else
					buffer.append ("			create Result.make_struct (<<%N")
				end
				l_first := True
				across fields as c loop
					if attached c.item.identifier as identifier then
						if l_first then
							l_first := False
						else
							buffer.append (",%N")
						end
						buffer.append ("					create {CE_STRUCTURE_FIELD}.make")
						if attached {AST_ARRAY_TYPE} c.item.type then
							buffer.append ("_arrayfield")
						elseif attached {AST_BITFIELD} c.item then
							buffer.append ("_bitfield")
						end
						buffer.append (" (%"" + identifier + "%"")
						if attached {AST_ARRAY_TYPE} c.item.type as a then
							buffer.append (", <<")
							from
								l_tmp_array := a
							until
								l_tmp_array = Void
							loop
								if l_tmp_array /= a then
									buffer.append (", ")
								end
								if l_tmp_array.dimension /= Void then
									buffer.append (l_tmp_array.dimension.eiffel_code)
								else
									-- flexible array member at end of struct; no dimension needed
									buffer.append ("0")
								end
								l_tmp_array ?= l_tmp_array.type
							end
							buffer.append (">>")
						end
						buffer.append (")")
					end
				end
				if not l_first then
					buffer.append ("%N")
				end
				buffer.append ("				>>)%N")

				across fields as c loop
					if
						attached c.item.gcc_attribute_specifier as gas and then
						attached gas.find_attribute ("__aligned__") as a and then
						a.expressions.count = 1
					then
						buffer.append ("			Result.field_table[%"" + c.item.identifier + "%"].alignment := " + a.expressions.first.eiffel_code + "%N")
					end
				end

				if
					attached node.gcc_attribute_specifier as gas and then
					attached gas.find_attribute ("__aligned__") as a and then
					a.expressions.count = 1
				then
					buffer.append ("			Result.alignment := " + a.expressions.first.eiffel_code + "%N")
				end

				if
					not node.oo_superclasses.is_empty and then
					across node.oo_superclasses as c some not c.item.source_info.is_pure_oo end
				then
					buffer.append ("			Result.set_superclasses (<<")
					l_first := True
					across node.oo_superclasses as c loop
						if not c.item.source_info.is_pure_oo then
							if l_first then
								l_first := False
							else
								buffer.append (", ")
							end
							buffer.append ("{" + eiffel_class_name (c.item.identifier) + "}")
						end
					end
					buffer.append (">>)%N")

					across node.oo_superclasses as c loop
						if not c.item.source_info.is_pure_oo then
							node.fields.start
							across c.item.fields as c2 loop
								if
									attached c2.item.identifier as super_identifier and then
									attached node.fields.item.identifier as sub_identifier and then
									not super_identifier.is_equal (sub_identifier)
								then
									buffer.append ("			Result.add_rename (%"" + super_identifier + "%", %"" + sub_identifier + "%")%N")
								end
								node.fields.forth
							end
						end
					end
				end

				buffer.append ("			init_structure_info%N")
				buffer.append ("		end%N%N")
			end
		end

	write_bitfield_helper
		local
			l_c_name: STRING_8
			l_tmp: BOOLEAN
		do
			if
				not node.source_info.is_pure_oo and then
				has_bitfields
			then
				if is_union then
					l_c_name := "union " + node.orig_c_identifier
				else
					l_c_name := "struct " + node.orig_c_identifier
				end
				buffer.append ("	bitfield_helper (a_operation: INTEGER_32; a_field_name, a_struct_pointer: POINTER; a_value: NATURAL_64): NATURAL_64%N")
				buffer.append ("		external%N")
				if node.source_info.is_project_source or node.orig_c_identifier.starts_with ("__anon") then
					buffer.append ("			%"C inline%"%N")
				else
					buffer.append ("			%"C inline use <" + node.source_info.source_file + ">%"%N")
				end
				buffer.append ("		alias%N")
				buffer.append ("			%"[%N")
				if node.source_info.is_project_source or node.orig_c_identifier.starts_with ("__anon") then
					buffer.append ("				" + l_c_name + " {%N")
					across node.fields as c loop
						buffer.append ("					" + c.item.c_code + ";%N")
					end
					buffer.append ("				};%N")
				end
				buffer.append ("				char *field_name = (char *)$a_field_name;%N")
				buffer.append ("				if($a_operation == 1) {%N")
				buffer.append ("					if(field_name == (char *)0) return sizeof(" + l_c_name + ");%N")
				l_tmp := False -- last was bitfield?
				across node.fields as c loop
					if attached {AST_BITFIELD} c.item then
						l_tmp := True
					elseif l_tmp and attached c.item.identifier as identifier then
						buffer.append ("					if(strcmp(field_name, %"" + identifier + "%") == 0) return offsetof(" + l_c_name + ", " + c.item.orig_c_identifier + ");%N")
						l_tmp := False
					end
				end
				buffer.append ("				} else {%N")
				buffer.append ("					" + l_c_name + " *p = (" + l_c_name + " *)$a_struct_pointer;%N")
				buffer.append ("					if($a_operation == 2) {%N")
				l_tmp := True -- first?
				across node.fields as c loop
					if attached {AST_BITFIELD} c.item and attached c.item.identifier as identifier then
						buffer.append ("						")
						if l_tmp then
							buffer.append ("     ")
							l_tmp := False
						else
							buffer.append ("else ")
						end
						buffer.append ("if(strcmp(field_name, %"" + identifier + "%") == 0) p->" + c.item.orig_c_identifier + " = (" + c.item.type.c_code + ")$a_value;%N")
					end
				end
				buffer.append ("					} else if($a_operation == 3) {%N")
				across node.fields as c loop
					if attached {AST_BITFIELD} c.item and attached c.item.identifier as identifier then
						buffer.append ("						if(strcmp(field_name, %"" + identifier + "%") == 0) return p->" + c.item.orig_c_identifier + ";%N")
					end
				end
				buffer.append ("					}%N")
				buffer.append ("				}%N")
				buffer.append ("			]%"%N")
				buffer.append ("		end%N%N")
			end
		end

feature {NONE}

	write_features (a_buffer: STRING_8; a_features: HASH_SET [AST_EIFFEL_FEATURE])
		local
			l_first: BOOLEAN
		do
			-- print public attributes
			l_first := True
			across a_features as c loop
				if
					attached {AST_VARIABLE_DECLARATION} c.item as vd and then
					vd.oo_class_assignment_data.applied_heuristic /= {OO_FUNCTIONS_TRANSFORMER}.Heuristic_loose_coupling
				then
					if l_first then
						a_buffer.append ("feature {ANY} -- public attributes%N%N")
						l_first := False
					end

					if vd.is_extern then
						a_buffer.append ("	")
						(create {NATIVE_FUNCTIONS_WRITER}).write_native_variable (vd, a_buffer)
						a_buffer.append ("%N%N")
					else
						write_attribute (a_buffer, vd)
					end
				end
			end

			-- print private attributes
			l_first := True
			across a_features as c loop
				if
					attached {AST_VARIABLE_DECLARATION} c.item as vd and then
					vd.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_loose_coupling
				then
					if l_first then
						a_buffer.append ("feature {" + node.identifier + "} -- private attributes%N%N")
						l_first := False
					end

					if vd.is_extern then
						a_buffer.append ("	")
						(create {NATIVE_FUNCTIONS_WRITER}).write_native_variable (vd, a_buffer)
						a_buffer.append ("%N%N")
					else
						write_attribute (a_buffer, vd)
					end
				end
			end

			-- print public routines
			l_first := True
			across a_features as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.oo_class_assignment_data.applied_heuristic /= {OO_FUNCTIONS_TRANSFORMER}.Heuristic_loose_coupling
				then
					if l_first then
						a_buffer.append ("feature {ANY} -- public routines%N%N")
						l_first := False
					end
					write_routine (a_buffer, fd)
				end
			end

			-- print private routines
			l_first := True
			across a_features as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_loose_coupling
				then
					if l_first then
						a_buffer.append ("feature {" + node.identifier)
						if
							across a_features as c2 some
								attached {AST_FUNCTION_DECLARATION} c2.item as fd2 and then
								fd2.oo_class_assignment_data.applied_heuristic = {OO_FUNCTIONS_TRANSFORMER}.Heuristic_loose_coupling and then
								fd2.needs_address
							end
						then
							a_buffer.append ("_STATIC")
						end
						a_buffer.append ("} -- private routines%N%N")
						l_first := False
					end
					write_routine (a_buffer, fd)
				end
			end
		end

	write_attribute (a_buffer: STRING_8; a_node: AST_VARIABLE_DECLARATION)
		require
			a_buffer /= Void
			a_node /= Void
		local
			l_eiffel_creator: AST_EIFFEL_CREATOR
			l_incomplete_initializer_info: like incomplete_initializer_info
			l_cur: AST_ARRAY_TYPE
		do
			a_buffer.append ("	" + a_node.identifier + ": CE_STATIC [" + a_node.type.eiffel_code + "]%N")
			a_buffer.append ("		once (%"PROCESS%")%N")
			if a_node.initializer /= Void then
				-- initializer; AST_EXPRESSION or (AST_INITIALIZER for arrays or structs)
				l_incomplete_initializer_info := incomplete_initializer_info (a_node.initializer)
				if
					not l_incomplete_initializer_info.main_array and
					not l_incomplete_initializer_info.subarrays and
					not has_self_reference_in_initializer (a_node)
				then
					create l_eiffel_creator.make
					l_eiffel_creator.indent := 3
					a_node.initializer.accept (l_eiffel_creator)
					a_buffer.append ("			Result.item := " + l_eiffel_creator.buffer + "%N")
				elseif
					not l_incomplete_initializer_info.subarrays and
					attached {AST_ARRAY_TYPE} a_node.type as array_type and
					attached {AST_INITIALIZER} a_node.initializer as initializer
				then
					a_buffer.append ("			Result.item := create {" + effective_type (array_type) + "}.make (<<" + array_type.dimension.eiffel_code + ">>)%N")
					across initializer.elements as cur_elem loop
						create l_eiffel_creator.make
						l_eiffel_creator.indent := 3
						cur_elem.item.accept (l_eiffel_creator)
						a_buffer.append ("			Result.item[" + (cur_elem.target_index - 1).out + "] := " + l_eiffel_creator.buffer + "%N")
					end
				else
					(create {EXCEPTIONS}).raise ("?!") -- shouldn't be possible
				end
			else
				if
					attached {AST_POINTER_TYPE} a_node.type or
					attached {AST_ARRAY_TYPE} a_node.type or
					attached {AST_STRUCT_TYPE} a_node.type
				then
					a_buffer.append ("			Result.item := create {")
					if
						(attached {AST_POINTER_TYPE} a_node.type as pt and then not attached {AST_FUNCTION_TYPE} pt.type) or
						attached {AST_ARRAY_TYPE} a_node.type
					then
						a_buffer.append (effective_type (a_node.type))
					else
						a_buffer.append (a_node.type.eiffel_code)
					end
					a_buffer.append ("}.make")
					if attached {AST_ARRAY_TYPE} a_node.type as array_type then
						a_buffer.append (" (<<")
						from
							l_cur := array_type
						until
							l_cur = Void
						loop
							if l_cur /= array_type then
								a_buffer.append (", ")
							end
							a_buffer.append (l_cur.dimension.eiffel_code)
							l_cur ?= l_cur.type
						end
						a_buffer.append (">>)")
					end
					a_buffer.append ("%N")
				end
			end
			a_buffer.append ("		end%N%N")
		end

	write_routine (a_buffer: STRING_8; a_node: AST_FUNCTION_DECLARATION)
		require
			a_buffer /= Void
			a_node /= Void
		do
			a_buffer.append (a_node.eiffel_code)
			a_buffer.append ("%N%N")
		end

	has_self_reference_in_initializer (a_node: AST_VARIABLE_DECLARATION): BOOLEAN
		require
			a_node /= Void and then a_node.initializer /= Void
		local
			l_result: CELL [BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			create l_result.put (False)
			create l_visitor
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_decl: AST_VARIABLE_DECLARATION; ag_result: CELL [BOOLEAN])
				do
					if ag_node.declaration_binding = ag_decl then
						ag_result.put (True)
					end
				end (?, a_node, l_result)
			a_node.accept (l_visitor)
			Result := l_result.item
		end

	incomplete_initializer_info (a_node: AST_EXPRESSION): TUPLE [main_array, subarrays: BOOLEAN]
		require
			a_node /= Void
		local
			l_const_eval: INTEGER_CONSTANT_EXPRESSION_EVALUATOR
			l_sub_info: like incomplete_initializer_info
		do
			create Result
			if
				attached {AST_INITIALIZER} a_node as initializer and then
				attached {AST_ARRAY_TYPE} initializer.type_binding as array_type
			then
				create l_const_eval.make (array_type.dimension)
				if l_const_eval.is_integer then
					Result.main_array := l_const_eval.integer /= initializer.elements.count
				else
					-- even possible?
					Result.main_array := l_const_eval.natural.to_integer_32 /= initializer.elements.count
				end

				across
					initializer.elements as c
				until
					Result.subarrays
				loop
					l_sub_info := incomplete_initializer_info (c.item)
					Result.subarrays := l_sub_info.main_array or l_sub_info.subarrays
				end
			end
		ensure
			Result /= Void
		end

	create_assigner_name (a_struct: AST_STRUCT_DECLARATION; a_field: AST_STRUCT_FIELD): STRING_8
		require
			a_struct /= Void
			a_field /= Void
			a_field.parent = a_struct
			a_field.identifier /= Void
		local
			l_routine_names: HASH_SET [STRING]
			l_i: NATURAL_32
		do
			create l_routine_names.make (16)
			across a_struct.oo_features as c loop
				l_routine_names.extend (c.item.identifier)
			end
			across a_struct.oo_static_features as c loop
				l_routine_names.extend (c.item.identifier)
			end

			from
				l_i := 1
				Result := "set_" + a_field.identifier
			until
				not l_routine_names.has (Result)
			loop
				l_i := l_i + 1
				Result := "set_" + a_field.identifier + l_i.out
			end
		ensure
			Result /= Void and then Result.starts_with ("set_" + a_field.identifier)
		end

invariant
	node /= Void

end
