class
	SWITCH_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_function_declaration,
			end_visit_function_declaration,
			visit_switch_statement,
			end_visit_switch_statement,
			end_visit_jump_statement
		end
	AST_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming switch statements"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			switch_index := 1
			Result := True
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		do
			if switch_index /= 1 then
				node.has_jump_statements := True
			end
		end

	visit_switch_statement (node: AST_SWITCH_STATEMENT): BOOLEAN
		do
			-- already add `done_state' so that `break' can reference it
			-- don't do the full transformation here because otherwise `break's can't find enclosing `switch'es anymore
			-- temporarily save `switch_index' in the jump state
			insert_adjacent_statement (node, create {AST_EIFFEL_JUMP_TARGET}.make_comment (switch_index, "switch " + switch_index.out + ", done:"), True)
			switch_index := switch_index + 1

			mark_requires_split (node)

			Result := True
		end

	end_visit_switch_statement (node: AST_SWITCH_STATEMENT)
		local
			branches: TWO_WAY_LIST [AST_STATEMENT]
			agent_visitor: AST_AGENT_VISITOR

			case_states: TWO_WAY_LIST [AST_EIFFEL_JUMP_TARGET]
			default_state: AST_EIFFEL_JUMP_TARGET
			done_state: AST_EIFFEL_JUMP_TARGET
			cur_switch_index: NATURAL_32
			switch_default: AST_SWITCH_DEFAULT
			eval: INTEGER_CONSTANT_EXPRESSION_EVALUATOR
			l_when: STRING_8
		do
			create branches.make
			create agent_visitor
			agent_visitor.agents.end_visit_switch_case := agent (ag_node: AST_SWITCH_CASE; ag_branches: TWO_WAY_LIST [AST_STATEMENT])
				do
					ag_branches.extend (ag_node)
					if
						not
						(attached {AST_COMPOUND_STATEMENT} ag_node.parent as cs and then
						attached {AST_SWITCH_STATEMENT} cs.parent)
					then
						mark_requires_split (ag_node)
					end
				end (?, branches)
			agent_visitor.agents.end_visit_switch_default := agent (ag_node: AST_SWITCH_DEFAULT; ag_branches: TWO_WAY_LIST [AST_STATEMENT])
				do
					ag_branches.extend (ag_node)
					if
						not
						(attached {AST_COMPOUND_STATEMENT} ag_node.parent as cs and then
						attached {AST_SWITCH_STATEMENT} cs.parent)
					then
						mark_requires_split (ag_node)
					end
				end (?, branches)
			node.statements.accept (agent_visitor)

			create case_states.make
			done_state ?= get_adjacent_statement (node, True)
			cur_switch_index := done_state.index
			done_state.index.set_item (0)

			-- create the head
			insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"inspect ", node.condition.ast_twin, " -- switch " + cur_switch_index.out + " head">>,
						create {AST_BASIC_TYPE}.make (Ast_bt_void)
					)), False)
			across branches as c loop
				if attached {AST_SWITCH_CASE} c.item as tmp then
					case_states.extend (create {AST_EIFFEL_JUMP_TARGET}.make_comment (0, "switch " + cur_switch_index.out + ", case " + tmp.condition.c_code))

					create eval.make (tmp.condition)
					if eval.is_integer then
						l_when := "when " + eval.integer.out + " then "
					else
						l_when := "when " + eval.natural.out + " then "
					end

					insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<l_when, create {AST_EIFFEL_JUMP}.make (case_states.last, Void)>>,
							create {AST_BASIC_TYPE}.make (Ast_bt_void)
						)), False)
					if not attached {AST_DECIMAL_CONSTANT} tmp.condition then
						({AST_EIFFEL_EXPRESSION} #? ({AST_EXPRESSION_STATEMENT} #? get_adjacent_statement (node, False)).expression).
								mixed.extend (" -- " + tmp.condition.eiffel_code)
					end
				elseif attached {AST_SWITCH_DEFAULT} c.item as tmp then
					switch_default := tmp
				end
			end
			if switch_default /= Void then
				create default_state.make_comment (0, "switch " + cur_switch_index.out + ", default:")
				insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"else ", create {AST_EIFFEL_JUMP}.make (default_state, Void)>>,
						create {AST_BASIC_TYPE}.make (Ast_bt_void)
					)), False)
			else
				insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"else ", create {AST_EIFFEL_JUMP}.make (done_state, Void)>>,
						create {AST_BASIC_TYPE}.make (Ast_bt_void)
					)), False)
			end
			insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make (
					"end",
					create {AST_BASIC_TYPE}.make (Ast_bt_void)
				)), False)

			-- add switch body
			across branches as c loop
				if attached {AST_SWITCH_CASE} c.item then
					replace_node (c.item, case_states.first)
					case_states.start
					case_states.remove
				elseif attached {AST_SWITCH_DEFAULT} c.item then
					replace_node (c.item, default_state)
				end
			end
			across node.statements.statements as c loop
				insert_adjacent_statement (node, c.item.ast_twin, False)
			end

			-- remove original switch statement
			delete_node (node)
		end

	end_visit_jump_statement (node: AST_JUMP_STATEMENT)
		local
			cur: AST_NODE
			eiffel_state: AST_EIFFEL_JUMP_TARGET
		do
			if node.type = C_kw_break then
				from
					cur := node.parent
				until
					cur = Void
				loop
					if attached {AST_SWITCH_STATEMENT} cur as tmp then
						eiffel_state ?= get_adjacent_statement (tmp, True)
						replace_node (node, create {AST_EIFFEL_JUMP}.make (eiffel_state, "switch " + eiffel_state.index.out + "; break"))
						cur := Void
					elseif attached {AST_WHILE_STATEMENT} cur then
						-- the `break' actually belongs to a while statement
						cur := Void
					else
						cur := cur.parent
					end
				end
			end
		end

feature {NONE}

	switch_index: NATURAL_32

	mark_requires_split (a_node: AST_NODE)
		local
			l_cur: AST_NODE
		do
			from
				l_cur := a_node.parent
			until
				attached {AST_FUNCTION_DECLARATION} l_cur
			loop
				if attached {AST_IF_STATEMENT} l_cur as tmp then
					tmp.requires_split := True
				elseif attached {AST_WHILE_STATEMENT} l_cur as tmp then
					tmp.requires_split := True
				end
				l_cur := l_cur.parent
			end
		end

end
