class
	AST_UTILITIES

inherit
	ANTLR_AST_TYPE_CONSTANTS

feature {ANY}

	replace_node (old_node, new_node: AST_NODE)
		require
			old_node /= Void and then old_node.parent /= Void
			new_node /= Void and then new_node.parent = Void
		local
--			tmp: STRING
			internal: INTERNAL
			parent: AST_NODE
			i: INTEGER
			field: ANY
		do
			if
				attached {AST_INITIALIZER} old_node.parent as initializer_parent and
				attached {AST_EXPRESSION} old_node as old_expression and
				attached {AST_EXPRESSION} new_node as new_expression
			then
				-- try to search from current position first; makes sense with visitor pattern
				initializer_parent.elements.search (old_expression)
				if initializer_parent.elements.after then
					-- not found? search from start
					initializer_parent.elements.start
					initializer_parent.elements.search (old_expression)
				end
				initializer_parent.elements.replace (new_expression)
				new_node.parent := old_node.parent
				old_node.parent := Void
			elseif
				attached {AST_GENERIC_EXPRESSION} old_node.parent as generic_expression_parent and
				attached {AST_EXPRESSION} old_node as old_expression and
				attached {AST_EXPRESSION} new_node as new_expression
			then
				if generic_expression_parent.right = old_expression then
					generic_expression_parent.right := new_expression
				elseif generic_expression_parent.left = old_expression then
					generic_expression_parent.left := new_expression
				end
			else
--				-- debug `replace_node' usage. should be done with `internal' as little as possible
--				tmp := old_node.parent.generator
--				if not replace_parents.has (tmp) then
--					replace_parents[tmp] := 1
--				else
--					replace_parents[tmp] := replace_parents[tmp] + 1
--				end
--				from
--					replace_parents.start
--				until
--					replace_parents.after
--				loop
--					Io.put_string (replace_parents.key_for_iteration + ": " + replace_parents.item_for_iteration.out + "%N")
--					replace_parents.forth
--				end
--				Io.put_string ("-----%N")

				-- generic replacement algorithm:
				create internal
				parent := old_node.parent
				from
					i := internal.field_count (parent)
				until
					i < 1
				loop
					field := internal.field (i, parent)
					if field = old_node then
						new_node.parent := parent
						old_node.parent := Void
						internal.set_reference_field (i, parent, new_node)
						i := 0
					elseif attached {LIST [ANY]}field as list then
						replace_expression_in_list (list, old_node, new_node)
					end
					i := i - 1
				end
			end
		ensure
			old_node.parent = Void
			new_node.parent = old old_node.parent
		end

	insert_adjacent_statement (node, new_node: AST_STATEMENT; after: BOOLEAN)
		require
			node /= Void and then node.parent /= Void
			new_node /= Void and then new_node.parent = Void
		local
			statement_list: TWO_WAY_LIST [AST_STATEMENT]
			tmp: AST_STATEMENT
		do
			if attached {AST_COMPOUND_STATEMENT} node.parent as compound_statement then
				statement_list := compound_statement.statements
				new_node.parent := compound_statement
			elseif attached {AST_LABELED_STATEMENT} node.parent as labeled_statement then
				if after then
					insert_adjacent_statement (labeled_statement, new_node, True)
				else
					tmp := labeled_statement.statement
					labeled_statement.statement := new_node
					insert_adjacent_statement (labeled_statement, tmp, True)
				end
			else
				(create {EXCEPTIONS}).raise ("[AST_UTILITIES.insert_adjacent_statement] unknown parent: " + node.parent.generator)
			end

			if statement_list /= Void then
				statement_list.start
				statement_list.search (node)
				if after then
					statement_list.put_right (new_node)
				else
					statement_list.put_left (new_node)
				end
			end
		ensure
			new_node.parent /= Void
		end

	get_adjacent_statement (node: AST_STATEMENT; next: BOOLEAN): AST_STATEMENT
		require
			node /= Void and then node.parent /= Void
		local
			statement_list: TWO_WAY_LIST [AST_STATEMENT]
		do
			if attached {AST_COMPOUND_STATEMENT} node.parent as compound_statement then
				statement_list := compound_statement.statements
			elseif attached {AST_LABELED_STATEMENT} node.parent as labeled_statement then
				Result := get_adjacent_statement (labeled_statement, next)
			else
				(create {EXCEPTIONS}).raise ("[AST_UTILITIES.get_adjacent_statement] unknown parent: " + node.parent.generator)
			end

			if statement_list /= Void then
				statement_list.start
				statement_list.search (node)
				if next then
					statement_list.forth
				else
					statement_list.back
				end
				if not statement_list.after and not statement_list.before then
					Result := statement_list.item
				end
			end
		end

	delete_node (node: AST_NODE)
			-- Remove the node from the parent (node keeps link to parent):
			-- o if the node is in a list of the parent, it's removed
			-- o if the node is an attribute, it's set to Void
		require
			node_exists: node /= Void
			node_has_parent: node.parent /= Void
		local
			internal: INTERNAL
			i, delete_index: INTEGER
			cur_field: ANY
			deleted: BOOLEAN
		do
			if attached {AST_LABELED_STATEMENT} node.parent as tmp then
				replace_node (node, create {AST_EXPRESSION_STATEMENT}.make (Void))
			else
				from
					create internal
					i := internal.field_count (node.parent)
				until
					i < 1 or deleted
				loop
					cur_field := internal.field (i, node.parent)
					if cur_field = node then
						internal.set_reference_field (i, node.parent, Void)
						deleted := True
					elseif
						attached {LIST [ANY]} cur_field as list and then
						(list.generating_type.generic_parameter_type (1) #? node) /= Void
					then
						delete_index := list.index_of (node, 1)
						if delete_index > 0 then
							list.go_i_th (delete_index)
							list.remove
							deleted := True
						end
					end
					i := i - 1
				end
				check node_deleted: deleted end
				node.parent := Void
			end
		ensure
			node.parent = Void
		end

	create_negated_expression (node: AST_EXPRESSION): AST_EXPRESSION
		require
			node /= Void
		local
			expression: AST_EXPRESSION
			new_boolean_constant: AST_BOOLEAN_CONSTANT
			new_type: INTEGER
		do
			-- remove parentheses
			from
				expression := node
			until
				not (attached {AST_PARENTHESIZED_EXPRESSION} expression as primary_expression and then primary_expression.expression /= Void)
			loop
				expression := primary_expression.expression
			end
			-- negate
			if attached {AST_BOOLEAN_CONSTANT}expression as tmp then
				new_boolean_constant := tmp.ast_twin
				if new_boolean_constant.value then
					new_boolean_constant.value := False
				else
					new_boolean_constant.value := True
				end
				Result := new_boolean_constant
			elseif attached {AST_UNARY_EXPRESSION} expression as tmp and then tmp.type = C_op_not then
				Result := tmp.expression.ast_twin
			elseif attached {AST_GENERIC_EXPRESSION} expression as tmp and then tmp.is_equality_expression then
				if tmp.type = C_op_equal then
					new_type := C_op_not_equal
				else
					new_type := C_op_equal
				end
				Result := create {AST_GENERIC_EXPRESSION}.make (new_type, tmp.left.ast_twin, tmp.right.ast_twin)
			else
				Result := create {AST_UNARY_EXPRESSION}.make (C_op_not, create {AST_PARENTHESIZED_EXPRESSION}.make (expression.ast_twin))
			end
		ensure
			Result /= Void and then Result.parent = Void
		end

	deref (type: AST_TYPE): AST_TYPE
		require
			attached {AST_POINTER_TYPE} type or attached {AST_ARRAY_TYPE} type
		do
			if attached {AST_POINTER_TYPE} type as tmp then
				Result := tmp.type
			elseif attached {AST_ARRAY_TYPE} type as tmp then
				Result := tmp.type
			end
		ensure
			Result /= Void
		end

	is_string_type (type: AST_TYPE): BOOLEAN
			-- checks for `char *' and `char []'
		do
			if
				(attached {AST_ARRAY_TYPE} type or attached {AST_POINTER_TYPE} type) and then
				attached {AST_BASIC_TYPE} deref (type) as basic_type and then
				basic_type.type = ast_bt_char
			then
				Result := True
			elseif attached {AST_EIFFEL_STRING_TYPE} type then
				Result := True
			else
				Result := False
			end
		end


feature {NONE}

--	replace_parents: HASH_TABLE [INTEGER, STRING]
--		once ("PROCESS")
--			create Result.make (16)
--		end

	replace_expression_in_list (list: LIST [ANY]; old_node, new_node: AST_NODE)
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > list.count
			loop
				if list.i_th (i) = old_node then
					new_node.parent := old_node.parent
					old_node.parent := Void
					list.put_i_th (new_node, i)
					i := list.count
				elseif attached {LIST [ANY]}list.i_th (i) as sub_list then
					replace_expression_in_list (sub_list, old_node, new_node)
				end
				i := i + 1
			end
		end

end
