class
	GRID_DIALOG

inherit
	EV_TITLED_WINDOW
		redefine
			initialize
		end
	VISION2_GC_WORKAROUND

create
	make

feature {NONE}

	make (a_title: STRING_8)
		do
			default_create
			protect_window_from_gc (Current)
			set_title (a_title)
		end

feature {ANY}

	grid: EV_GRID

	set_columns (titles: ARRAY [STRING_8])
		require
			titles /= Void and then not titles.is_empty
			grid.column_count = 0
		do
			grid.set_column_count_to (titles.count)
			across titles as c loop
				grid.column (c.target_index).set_title (c.item)
			end
		ensure
			grid.column_count = titles.count
		end

	resize_columns_to_content
		local
			l_tmp_row_index: INTEGER_32
		do
			l_tmp_row_index := grid.row_count + 1
			across 1 |..| grid.column_count as i loop
				grid.set_item (i.item, l_tmp_row_index, create {EV_GRID_LABEL_ITEM}.make_with_text (grid.column (i.item).title))
				grid.column (i.item).resize_to_content
				grid.column (i.item).set_width (grid.column (i.item).width + 11)
			end
			grid.remove_row (l_tmp_row_index)
		end

feature {NONE}

	ctrl_down: BOOLEAN

feature {NONE}

	initialize
		local
			l_main_container: EV_VERTICAL_BOX
		do
			Precursor {EV_TITLED_WINDOW}

			create l_main_container
			extend (l_main_container)

			create grid
			grid.mouse_wheel_actions.extend (agent mouse_wheeled)
			l_main_container.extend (grid)

			key_press_actions.extend (agent key_released (?, True))
			key_release_actions.extend (agent key_released (?, False))

			close_request_actions.extend (agent close_request)
			set_size (800, 600)
		end

	mouse_wheeled (a_direction: INTEGER_32)
		do
			if a_direction = 1 then
				grid.vertical_scroll_bar.step_backward
			else
				grid.vertical_scroll_bar.step_forward
			end
		end

	key_released (a_key: EV_KEY; a_press: BOOLEAN)
		local
			l_find_dialog: EV_INPUT_DIALOG
		do
			if a_press then
				if a_key.code = {EV_KEY_CONSTANTS}.Key_ctrl then
					ctrl_down := True
				elseif ctrl_down and a_key.code = {EV_KEY_CONSTANTS}.Key_f then
					create l_find_dialog.make_with_text ("Find what:")
					l_find_dialog.set_title ("Find")
					l_find_dialog.ok_select_actions.extend (agent (ag_find_dialog: EV_INPUT_DIALOG)
						local
							l_s: STRING_8
							l_information_dialog: EV_INFORMATION_DIALOG
						do
							l_s := ag_find_dialog.input
							l_s.left_adjust
							l_s.right_adjust
							if
								not l_s.is_empty and then
								not select_item (l_s)
							then
								create l_information_dialog.make_with_text ("Search string '" + l_s + "' not found!")
								l_information_dialog.set_title ("Not found")
								l_information_dialog.show_modal_to_window (Current)
							end
						end (l_find_dialog))
					l_find_dialog.show_modal_to_window (Current)
				end
			else
				if a_key.code = {EV_KEY_CONSTANTS}.Key_ctrl then
					ctrl_down := False
				end
			end
		end

	close_request
		do
			destroy
		end

	select_item (a_text: STRING_8): BOOLEAN
		require
			a_text /= Void and then not a_text.is_empty
		local
			l_row, l_column: INTEGER_32
			l_tmp: EV_GRID_ROW
		do
			from
				l_row := 1
			until
				Result or l_row > grid.row_count
			loop
				from
					l_column := 1
				until
					Result or l_column > grid.column_count
				loop
					if
						attached {EV_GRID_LABEL_ITEM} grid.item (l_column, l_row) as gli and then
						gli.text.has_substring (a_text)
					then
						from
							l_tmp := grid.row (l_row).parent_row
						until
							l_tmp = Void
						loop
							l_tmp.expand
							l_tmp := l_tmp.parent_row
						end
						grid.row (l_row).ensure_visible
						grid.select_row (l_row)
						Result := True
					end
					l_column := l_column + 1
				end
				l_row := l_row + 1
			end
		end

end
